/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Alexander Bokovoy <alexander.bokovoy@nokia.com>
**
** This file is part of the Quill Image Filters package.
**
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/

#include "readerstorage.h"
#include <QByteArray>
#include <QBuffer>
#include <QImageReader>
#include <QFile>

const QByteArray ReaderStorage::JpegFormatShort("jpg");
const QByteArray ReaderStorage::JpegFormatLong("jpeg");

ReaderStorageItem::ReaderStorageItem() :
    m_ByteArray(new QByteArray()),
    m_Buffer(new QBuffer()),
    m_reader(0),
    filterError(QuillImageFilter::NoError)
{
    m_Buffer->setBuffer(m_ByteArray);
}

ReaderStorageItem::~ReaderStorageItem()
{
    if(m_Buffer) {
        delete m_Buffer;
    }
    m_Buffer = NULL;
    if(m_ByteArray) {
        delete m_ByteArray;
    }
    m_ByteArray = NULL;
    if(m_reader){
        delete m_reader;
    }
    m_reader = NULL;
}

QBuffer* ReaderStorageItem::buffer()
{
    return m_Buffer;
}

QByteArray* ReaderStorageItem::byteArray()
{
    return m_ByteArray;
}

QImageReader* ReaderStorageItem::reader()
{
    return m_reader;
}

void ReaderStorageItem::setImageReader(QImageReader* imageReader)
{
    // QImageReader* gets allocated in loadStorageItemFromFile
    // We keep the allocated memory here.
    // LoadFilter dtor calls discardReader which eventually destroys
    // ReaderStorageItem
    m_reader = imageReader;
}

void ReaderStorageItem::setError(QuillImageFilter::ImageFilterError error)
{
    filterError = error;
}

QuillImageFilter::ImageFilterError ReaderStorageItem::error()
{
    return filterError;
}


ReaderStorage::ReaderStorage()
{
    m_ReaderCache.setMaxCost(1);
}

void ReaderStorage::discardReader(int uniqueKey)
{
    // Deletes the object associated with key key.
    m_ReaderCache.remove(uniqueKey);
}

bool ReaderStorage::isJpeg(const QByteArray &format) {
    return (format == JpegFormatShort) || (format == JpegFormatLong);
}

QImageReader* ReaderStorage::createImageReader(int uniqueKey, const QString& fileName,
                                               const QByteArray& format)
{
   ReaderStorageItem* readerStorageItem = (m_ReaderCache.object(uniqueKey));

   // With formats other than jpeg, using the same filter twice might
   // not be safe, thus forcing recreation in these cases.
   if (!readerStorageItem || !readerStorageItem->reader() ||
       !isJpeg(readerStorageItem->reader()->format()))
       {
       // setMaxCost of m_ReaderCache is one. This make sure that readerStorageItem
       // is always the current valid one. ReaderStorage::discardReader deletes this
       // this readerStorageItem. LoadFilter destructor calls discardReader to cleanup
       // the allocated memory.
       readerStorageItem = new ReaderStorageItem();
       m_ReaderCache.insert(uniqueKey,readerStorageItem);
       if( !loadStorageItemFromFile(fileName,readerStorageItem,format) )
           return NULL;
   }
   else {
       // Reader does not work correctly for the same file twice.
       readerStorageItem->buffer()->seek(0);
   }
   return readerStorageItem->reader();
}

QImageReader* ReaderStorage::getReaderFromStorage(int uniqueKey)
{
   ReaderStorageItem* readerStorageItem = (m_ReaderCache.object(uniqueKey));
   if( readerStorageItem ) {
        return readerStorageItem->reader();
   }
   return NULL;
}

bool ReaderStorage::loadStorageItemFromFile(const QString& fileName,
                                            ReaderStorageItem* readerStorageItem,
                                            const QByteArray& format)
{
    QFile file(fileName);

    bool returnValue;
    if(file.open(QIODevice::ReadOnly)) {
        QByteArray *fileByte = readerStorageItem->byteArray();
        *fileByte = file.readAll();
        if(fileByte->isEmpty()) {
            readerStorageItem->setError(QuillImageFilter::FileNotFoundError);
            returnValue = false;
        }
        else {
            // Loading data from file to QImageReader and putting it to
            // the ReaderStorageItem
            QBuffer* buffer = readerStorageItem->buffer();
            buffer->setBuffer(fileByte);
            buffer->open(QIODevice::ReadOnly);
            readerStorageItem->setImageReader(new QImageReader(buffer, format));
            returnValue = true;
        }
    }
    else {
        if(!file.exists())
            readerStorageItem->setError(QuillImageFilter::FileNotFoundError);
        else
            readerStorageItem->setError(QuillImageFilter::FileCanNotOpenError);
        returnValue = false;
    }
    return returnValue;
}

QuillImageFilter::ImageFilterError ReaderStorage::error(int uniqueKey)
{
    ReaderStorageItem* readerStorageItem = (m_ReaderCache.object(uniqueKey));

    if( readerStorageItem ) {
         return readerStorageItem->error();
    }
    return QuillImageFilter::UnspecifiedError;
}
