/****************************************************************************
**
** Copyright (C) 2009 Nokia Corporation and/or its subsidiary(-ies).
** Contact: Alexander Bokovoy <alexander.bokovoy@nokia.com>
**
** This file is part of the Quill Image Filters package.
**
** Commercial Usage
** Licensees holding valid Qt Commercial licenses may use this file in
** accordance with the Qt Commercial License Agreement provided with the
** Software or, alternatively, in accordance with the terms contained in
** a written agreement between you and Nokia.
**
** GNU Lesser General Public License Usage
** Alternatively, this file may be used under the terms of the GNU Lesser
** General Public License version 2.1 as published by the Free Software
** Foundation and appearing in the file LICENSE.LGPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU Lesser General Public License version 2.1 requirements
** will be met: http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
**
** In addition, as a special exception, Nokia gives you certain
** additional rights. These rights are described in the Nokia Qt LGPL
** Exception version 1.0, included in the file LGPL_EXCEPTION.txt in this
** package.
**
** GNU General Public License Usage
** Alternatively, this file may be used under the terms of the GNU
** General Public License version 3.0 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.  Please review the following information to
** ensure the GNU General Public License version 3.0 requirements will be
** met: http://www.gnu.org/copyleft/gpl.html.
**
** If you are unsure which license is appropriate for your use, please
** contact the sales department at qt-sales@nokia.com.
**
****************************************************************************/

/*!
  \class ReaderStorage

  \brief Container for ReaderStorageItem

  ReaderStorageItem contains image buffer which is used by LoadFilter.
  Image buffer is created on demand and releases when it's not needed.
  Each LoadFilter will have unique ReaderStorageItem associated with
  integer key. So there would be similar files with different unique
  keys.
*/

#ifndef READERSTORAGE_H
#define READERSTORAGE_H

#include <QCache>
#include <QLatin1String>
#include "quillimagefilter.h"

class QByteArray;
class QBuffer;
class QImageReader;

class ReaderStorageItem
{
public:
    ReaderStorageItem();
    ~ReaderStorageItem();

    inline QBuffer* buffer();
    inline QByteArray* byteArray();
    inline QImageReader* reader();
    inline void setImageReader(QImageReader* imageReader);
    inline QuillImageFilter::ImageFilterError error();
    void setError(QuillImageFilter::ImageFilterError error);

private:
    QByteArray                         *m_ByteArray;
    QBuffer                            *m_Buffer;
    QImageReader                       *m_reader;
    QuillImageFilter::ImageFilterError filterError;
};

class ReaderStorage
{
public:
    ReaderStorage();
    //! Creates ReaderStorageItem by loading the file data into QImageReader.
    //! After creating the storage item, it's inserted into the QCache container.
    //!
    //! \param uniqueKey, It's a unique key for each ReaderStorageItem instance.
    //! \param file which will be loaded into the buffer.
    //! \param MimeType of the given file.
    //! \return created QImageReader instance
    QImageReader* createImageReader(int uniqueKey, const QString& fileName, const QByteArray& format);

    //! Helper function to detect the jpeg format
    static bool isJpeg(const QByteArray &format);

    //! Returns the QImageReader from the ReaderStorageItem associted with the key.
    //!
    //! \param uniqueKey, It's a unique key for each ReaderStorageItem instance.
    //! \return QImageReader instance which was previously loaded.
    QImageReader* getReaderFromStorage(int uniqueKey);

    //! Accessor method for error code. Errocode is associated with each ReaderStorageItem.
    //!
    //! \param uniqueKey, It's a unique key for each ReaderStorageItem instance.
    //! \return QuillImageFilter error code.
    QuillImageFilter::ImageFilterError error(int uniqueKey);

    //! Discards the ReaderStorageItem from the QCache container.
    //! \param uniqueKey, It's a unique key for each ReaderStorageItem instance.
    void discardReader(int uniqueKey);

private:
    bool loadStorageItemFromFile(const QString& fileName, ReaderStorageItem* readerStorageItem, const QByteArray& format);

private:
    QCache<int,ReaderStorageItem> m_ReaderCache;

    static const QByteArray JpegFormatShort;
    static const QByteArray JpegFormatLong;
};

#endif // READERSTORAGE_H
